@@ -1,3 +1,5 @@
+// SPDX-License-Identifier: Apache-2.0
+// 
 // Copyright 2008-2016 Conrad Sanderson (http://conradsanderson.id.au)
 // Copyright 2008-2016 National ICT Australia (NICTA)
 // 
@@ -27,7 +29,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const T1& >::result
 real(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return X;
   }
@@ -40,7 +42,7 @@ arma_inline
 const T1&
 real(const BaseCube<typename T1::pod_type, T1>& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return X.get_ref();
   }
@@ -50,12 +52,12 @@ real(const BaseCube<typename T1::pod_type, T1>& X)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const T1&
-real(const SpBase<typename T1::pod_type,T1>& A)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::no), const T1& >::result
+real(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return A.get_ref();
+  return X;
   }
 
 
@@ -66,7 +68,7 @@ inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtOp<typename T1::pod_type, T1, op_real> >::result
 real(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return mtOp<typename T1::pod_type, T1, op_real>( X );
   }
@@ -79,7 +81,7 @@ inline
 const mtOpCube<typename T1::pod_type, T1, op_real>
 real(const BaseCube<std::complex<typename T1::pod_type>, T1>& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return mtOpCube<typename T1::pod_type, T1, op_real>( X.get_ref() );
   }
@@ -89,12 +91,12 @@ real(const BaseCube<std::complex<typename T1::pod_type>, T1>& X)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const mtSpOp<typename T1::pod_type, T1, spop_real>
-real(const SpBase<std::complex<typename T1::pod_type>,T1>& A)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtSpOp<typename T1::pod_type, T1, spop_real> >::result
+real(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return mtSpOp<typename T1::pod_type, T1, spop_real>(A.get_ref());
+  return mtSpOp<typename T1::pod_type, T1, spop_real>(X);
   }
 
 
@@ -105,55 +107,10 @@ real(const SpBase<std::complex<typename T1::pod_type>,T1>& A)
 template<typename T1>
 arma_warn_unused
 inline
-const Gen< Mat<typename T1::pod_type>, gen_zeros >
-imag(const Base<typename T1::pod_type,T1>& X)
-  {
-  arma_extra_debug_sigprint();
-  
-  const Proxy<T1> A(X.get_ref());
-  
-  return Gen< Mat<typename T1::pod_type>, gen_zeros>(A.get_n_rows(), A.get_n_cols());
-  }
-
-
-
-template<typename T1>
-arma_warn_unused
-inline
-const GenCube<typename T1::pod_type, gen_zeros>
-imag(const BaseCube<typename T1::pod_type,T1>& X)
-  {
-  arma_extra_debug_sigprint();
-  
-  const ProxyCube<T1> A(X.get_ref());
-  
-  return GenCube<typename T1::pod_type, gen_zeros>(A.get_n_rows(), A.get_n_cols(), A.get_n_slices());
-  }
-
-
-
-template<typename T1>
-arma_warn_unused
-inline
-SpMat<typename T1::pod_type>
-imag(const SpBase<typename T1::pod_type,T1>& A)
-  {
-  arma_extra_debug_sigprint();
-  
-  const SpProxy<T1> P(A.get_ref());
-  
-  return SpMat<typename T1::pod_type>(P.get_n_rows(), P.get_n_cols());
-  }
-
-
-
-template<typename T1>
-arma_warn_unused
-inline
-typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtOp<typename T1::pod_type, T1, op_imag> >::result
+typename enable_if2< is_arma_type<T1>::value, const mtOp<typename T1::pod_type, T1, op_imag> >::result
 imag(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return mtOp<typename T1::pod_type, T1, op_imag>( X );
   }
@@ -164,9 +121,9 @@ template<typename T1>
 arma_warn_unused
 inline
 const mtOpCube<typename T1::pod_type, T1, op_imag>
-imag(const BaseCube<std::complex<typename T1::pod_type>,T1>& X)
+imag(const BaseCube<typename T1::elem_type,T1>& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return mtOpCube<typename T1::pod_type, T1, op_imag>( X.get_ref() );
   }
@@ -176,12 +133,12 @@ imag(const BaseCube<std::complex<typename T1::pod_type>,T1>& X)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const mtSpOp<typename T1::pod_type, T1, spop_imag>
-imag(const SpBase<std::complex<typename T1::pod_type>,T1>& A)
+typename enable_if2< is_arma_sparse_type<T1>::value, const mtSpOp<typename T1::pod_type, T1, spop_imag> >::result
+imag(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return mtSpOp<typename T1::pod_type, T1, spop_imag>(A.get_ref());
+  return mtSpOp<typename T1::pod_type, T1, spop_imag>(X);
   }
 
 
@@ -195,7 +152,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_log> >::result
 log(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_log>(A);
   }
@@ -208,7 +165,7 @@ arma_inline
 const eOpCube<T1, eop_log>
 log(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_log>(A.get_ref());
   }
@@ -224,7 +181,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_log2> >::result
 log2(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_log2>(A);
   }
@@ -237,7 +194,7 @@ arma_inline
 const eOpCube<T1, eop_log2>
 log2(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_log2>(A.get_ref());
   }
@@ -253,7 +210,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_log10> >::result
 log10(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_log10>(A);
   }
@@ -266,7 +223,7 @@ arma_inline
 const eOpCube<T1, eop_log10>
 log10(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_log10>(A.get_ref());
   }
@@ -282,7 +239,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_log1p> >::result
 log1p(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_log1p>(A);
   }
@@ -295,7 +252,7 @@ arma_inline
 typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_log1p> >::result
 log1p(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_log1p>(A.get_ref());
   }
@@ -311,7 +268,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_exp> >::result
 exp(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_exp>(A);
   }
@@ -324,7 +281,7 @@ arma_inline
 const eOpCube<T1, eop_exp>
 exp(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_exp>(A.get_ref());
   }
@@ -339,7 +296,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_exp2> >::result
 exp2(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_exp2>(A);
   }
@@ -352,7 +309,7 @@ arma_inline
 const eOpCube<T1, eop_exp2>
 exp2(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_exp2>(A.get_ref());
   }
@@ -367,7 +324,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_exp10> >::result
 exp10(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_exp10>(A);
   }
@@ -380,7 +337,7 @@ arma_inline
 const eOpCube<T1, eop_exp10>
 exp10(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_exp10>(A.get_ref());
   }
@@ -395,7 +352,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_expm1> >::result
 expm1(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_expm1>(A);
   }
@@ -408,7 +365,7 @@ arma_inline
 typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_expm1> >::result
 expm1(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_expm1>(A.get_ref());
   }
@@ -425,7 +382,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_abs> >::result
 abs(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_abs>(X);
   }
@@ -438,7 +395,7 @@ arma_inline
 const eOpCube<T1, eop_abs>
 abs(const BaseCube<typename T1::elem_type,T1>& X, const typename arma_not_cx<typename T1::elem_type>::result* junk = nullptr)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   arma_ignore(junk);
   
@@ -453,7 +410,7 @@ inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtOp<typename T1::pod_type, T1, op_abs> >::result
 abs(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return mtOp<typename T1::pod_type, T1, op_abs>(X);
   }
@@ -466,7 +423,7 @@ inline
 const mtOpCube<typename T1::pod_type, T1, op_abs>
 abs(const BaseCube< std::complex<typename T1::pod_type>,T1>& X, const typename arma_cx_only<typename T1::elem_type>::result* junk = nullptr)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   arma_ignore(junk);
   
@@ -478,27 +435,26 @@ abs(const BaseCube< std::complex<typename T1::pod_type>,T1>& X, const typename a
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_abs>
-abs(const SpBase<typename T1::elem_type,T1>& X, const typename arma_not_cx<typename T1::elem_type>::result* junk = nullptr)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::no), const SpOp<T1, spop_abs> >::result
+abs(const T1& X)
   {
-  arma_extra_debug_sigprint();
-  arma_ignore(junk);
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_abs>(X.get_ref());
+  return SpOp<T1, spop_abs>(X);
   }
 
 
 
+
 template<typename T1>
 arma_warn_unused
 arma_inline
-const mtSpOp<typename T1::pod_type, T1, spop_cx_abs>
-abs(const SpBase< std::complex<typename T1::pod_type>, T1>& X, const typename arma_cx_only<typename T1::elem_type>::result* junk = nullptr)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtSpOp<typename T1::pod_type, T1, spop_cx_abs> >::result
+abs(const T1& X)
   {
-  arma_extra_debug_sigprint();
-  arma_ignore(junk);
+  arma_debug_sigprint();
   
-  return mtSpOp<typename T1::pod_type, T1, spop_cx_abs>(X.get_ref());
+  return mtSpOp<typename T1::pod_type, T1, spop_cx_abs>(X);
   }
 
 
@@ -513,7 +469,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_arg> >::result
 arg(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_arg>(X);
   }
@@ -526,7 +482,7 @@ arma_inline
 const eOpCube<T1, eop_arg>
 arg(const BaseCube<typename T1::elem_type,T1>& X, const typename arma_not_cx<typename T1::elem_type>::result* junk = nullptr)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   arma_ignore(junk);
   
@@ -541,7 +497,7 @@ inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtOp<typename T1::pod_type, T1, op_arg> >::result
 arg(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return mtOp<typename T1::pod_type, T1, op_arg>(X);
   }
@@ -554,7 +510,7 @@ inline
 const mtOpCube<typename T1::pod_type, T1, op_arg>
 arg(const BaseCube< std::complex<typename T1::pod_type>,T1>& X, const typename arma_cx_only<typename T1::elem_type>::result* junk = nullptr)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   arma_ignore(junk);
   
@@ -566,13 +522,12 @@ arg(const BaseCube< std::complex<typename T1::pod_type>,T1>& X, const typename a
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_arg>
-arg(const SpBase<typename T1::elem_type,T1>& X, const typename arma_not_cx<typename T1::elem_type>::result* junk = nullptr)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::no), const SpOp<T1, spop_arg> >::result
+arg(const T1& X)
   {
-  arma_extra_debug_sigprint();
-  arma_ignore(junk);
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_arg>(X.get_ref());
+  return SpOp<T1, spop_arg>(X);
   }
 
 
@@ -580,13 +535,12 @@ arg(const SpBase<typename T1::elem_type,T1>& X, const typename arma_not_cx<typen
 template<typename T1>
 arma_warn_unused
 arma_inline
-const mtSpOp<typename T1::pod_type, T1, spop_cx_arg>
-arg(const SpBase< std::complex<typename T1::pod_type>, T1>& X, const typename arma_cx_only<typename T1::elem_type>::result* junk = nullptr)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::yes), const mtSpOp<typename T1::pod_type, T1, spop_cx_arg> >::result
+arg(const T1& X)
   {
-  arma_extra_debug_sigprint();
-  arma_ignore(junk);
+  arma_debug_sigprint();
   
-  return mtSpOp<typename T1::pod_type, T1, spop_cx_arg>(X.get_ref());
+  return mtSpOp<typename T1::pod_type, T1, spop_cx_arg>(X);
   }
 
 
@@ -600,7 +554,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_square> >::result
 square(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_square>(A);
   }
@@ -613,7 +567,7 @@ arma_inline
 const eOpCube<T1, eop_square>
 square(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_square>(A.get_ref());
   }
@@ -623,12 +577,12 @@ square(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_square>
-square(const SpBase<typename T1::elem_type,T1>& A)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_square> >::result
+square(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_square>(A.get_ref());
+  return SpOp<T1, spop_square>(X);
   }
 
 
@@ -642,7 +596,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_sqrt> >::result
 sqrt(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_sqrt>(A);
   }
@@ -655,7 +609,7 @@ arma_inline
 const eOpCube<T1, eop_sqrt>
 sqrt(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_sqrt>(A.get_ref());
   }
@@ -665,12 +619,54 @@ sqrt(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_sqrt>
-sqrt(const SpBase<typename T1::elem_type,T1>& A)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_sqrt> >::result
+sqrt(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_sqrt>(A.get_ref());
+  return SpOp<T1, spop_sqrt>(X);
+  }
+
+
+
+//
+// cbrt
+
+template<typename T1>
+arma_warn_unused
+arma_inline
+typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_cbrt> >::result
+cbrt(const T1& A)
+  {
+  arma_debug_sigprint();
+  
+  return eOp<T1, eop_cbrt>(A);
+  }
+
+
+
+template<typename T1>
+arma_warn_unused
+arma_inline
+typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_cbrt> >::result
+cbrt(const BaseCube<typename T1::elem_type,T1>& A)
+  {
+  arma_debug_sigprint();
+  
+  return eOpCube<T1, eop_cbrt>(A.get_ref());
+  }
+
+
+
+template<typename T1>
+arma_warn_unused
+arma_inline
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::no), const SpOp<T1, spop_cbrt> >::result
+cbrt(const T1& X)
+  {
+  arma_debug_sigprint();
+  
+  return SpOp<T1, spop_cbrt>(X);
   }
 
 
@@ -681,12 +677,12 @@ sqrt(const SpBase<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const T1&
-conj(const Base<typename T1::pod_type,T1>& A)
+typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const T1& >::result
+conj(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return A.get_ref();
+  return X;
   }
 
 
@@ -697,7 +693,7 @@ arma_inline
 const T1&
 conj(const BaseCube<typename T1::pod_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return A.get_ref();
   }
@@ -707,12 +703,12 @@ conj(const BaseCube<typename T1::pod_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const T1&
-conj(const SpBase<typename T1::pod_type,T1>& A)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::no), const T1& >::result
+conj(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return A.get_ref();
+  return X;
   }
 
 
@@ -720,12 +716,12 @@ conj(const SpBase<typename T1::pod_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const eOp<T1, eop_conj>
-conj(const Base<std::complex<typename T1::pod_type>,T1>& A)
+typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::yes), const eOp<T1, eop_conj> >::result
+conj(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return eOp<T1, eop_conj>(A.get_ref());
+  return eOp<T1, eop_conj>(A);
   }
 
 
@@ -736,7 +732,7 @@ arma_inline
 const eOpCube<T1, eop_conj>
 conj(const BaseCube<std::complex<typename T1::pod_type>,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
 
   return eOpCube<T1, eop_conj>(A.get_ref());
   }
@@ -746,12 +742,12 @@ conj(const BaseCube<std::complex<typename T1::pod_type>,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_conj>
-conj(const SpBase<std::complex<typename T1::pod_type>,T1>& A)
+typename enable_if2< (is_arma_sparse_type<T1>::value && is_cx<typename T1::elem_type>::yes), const SpOp<T1, spop_conj> >::result
+conj(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_conj>(A.get_ref());
+  return SpOp<T1, spop_conj>(X);
   }
 
 
@@ -761,12 +757,12 @@ conj(const SpBase<std::complex<typename T1::pod_type>,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const eOp<T1, eop_pow>
-pow(const Base<typename T1::elem_type,T1>& A, const typename T1::elem_type exponent)
+typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_pow> >::result
+pow(const T1& A, const typename T1::elem_type exponent)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return eOp<T1, eop_pow>(A.get_ref(), exponent);
+  return eOp<T1, eop_pow>(A, exponent);
   }
 
 
@@ -777,7 +773,7 @@ arma_inline
 const eOpCube<T1, eop_pow>
 pow(const BaseCube<typename T1::elem_type,T1>& A, const typename T1::elem_type exponent)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_pow>(A.get_ref(), exponent);
   }
@@ -789,14 +785,14 @@ pow(const BaseCube<typename T1::elem_type,T1>& A, const typename T1::elem_type e
 template<typename T1>
 arma_warn_unused
 arma_inline
-const eOp<T1, eop_pow>
-pow(const Base<typename T1::elem_type,T1>& A, const typename T1::elem_type::value_type exponent)
+typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::yes), const eOp<T1, eop_pow> >::result
+pow(const T1& A, const typename T1::elem_type::value_type exponent)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::elem_type eT;
   
-  return eOp<T1, eop_pow>(A.get_ref(), eT(exponent));
+  return eOp<T1, eop_pow>(A, eT(exponent));
   }
 
 
@@ -805,11 +801,11 @@ template<typename T1>
 arma_warn_unused
 arma_inline
 const eOpCube<T1, eop_pow>
-pow(const BaseCube<typename T1::elem_type,T1>& A, const typename T1::elem_type::value_type exponent)
+pow(const BaseCube<std::complex<typename T1::pod_type>,T1>& A, const typename T1::elem_type::value_type exponent)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  typedef typename T1::elem_type eT;
+  typedef std::complex<typename T1::pod_type> eT;
   
   return eOpCube<T1, eop_pow>(A.get_ref(), eT(exponent));
   }
@@ -825,7 +821,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_floor> >::result
 floor(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_floor>(A);
   }
@@ -838,7 +834,7 @@ arma_inline
 const eOpCube<T1, eop_floor>
 floor(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_floor>(A.get_ref());
   }
@@ -848,12 +844,12 @@ floor(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_floor>
-floor(const SpBase<typename T1::elem_type,T1>& X)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_floor> >::result
+floor(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_floor>(X.get_ref());
+  return SpOp<T1, spop_floor>(X);
   }
 
 
@@ -867,7 +863,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_ceil> >::result
 ceil(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_ceil>(A);
   }
@@ -880,7 +876,7 @@ arma_inline
 const eOpCube<T1, eop_ceil>
 ceil(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_ceil>(A.get_ref());
   }
@@ -890,12 +886,12 @@ ceil(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_ceil>
-ceil(const SpBase<typename T1::elem_type,T1>& X)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_ceil> >::result
+ceil(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_ceil>(X.get_ref());
+  return SpOp<T1, spop_ceil>(X);
   }
 
 
@@ -909,7 +905,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_round> >::result
 round(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_round>(A);
   }
@@ -922,7 +918,7 @@ arma_inline
 const eOpCube<T1, eop_round>
 round(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_round>(A.get_ref());
   }
@@ -932,12 +928,12 @@ round(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_round>
-round(const SpBase<typename T1::elem_type,T1>& X)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_round> >::result
+round(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_round>(X.get_ref());
+  return SpOp<T1, spop_round>(X);
   }
 
 
@@ -951,7 +947,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_trunc> >::result
 trunc(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_trunc>(A);
   }
@@ -964,7 +960,7 @@ arma_inline
 const eOpCube<T1, eop_trunc>
 trunc(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_trunc>(A.get_ref());
   }
@@ -974,12 +970,12 @@ trunc(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_trunc>
-trunc(const SpBase<typename T1::elem_type,T1>& X)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_trunc> >::result
+trunc(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_trunc>(X.get_ref());
+  return SpOp<T1, spop_trunc>(X);
   }
 
 
@@ -993,7 +989,7 @@ arma_inline
 typename arma_scalar_only<eT>::result
 sign(const eT x)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return arma_sign(x);
   }
@@ -1006,7 +1002,7 @@ arma_inline
 typename enable_if2< is_arma_type<T1>::value, const eOp<T1, eop_sign> >::result
 sign(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_sign>(A);
   }
@@ -1019,7 +1015,7 @@ arma_inline
 const eOpCube<T1, eop_sign>
 sign(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_sign>(A.get_ref());
   }
@@ -1029,12 +1025,12 @@ sign(const BaseCube<typename T1::elem_type,T1>& A)
 template<typename T1>
 arma_warn_unused
 arma_inline
-const SpOp<T1, spop_sign>
-sign(const SpBase<typename T1::elem_type,T1>& X)
+typename enable_if2< is_arma_sparse_type<T1>::value, const SpOp<T1, spop_sign> >::result
+sign(const T1& X)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  return SpOp<T1, spop_sign>(X.get_ref());
+  return SpOp<T1, spop_sign>(X);
   }
 
 
@@ -1048,7 +1044,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_erf> >::result
 erf(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_erf>(A);
   }
@@ -1061,7 +1057,7 @@ arma_inline
 typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_erf> >::result
 erf(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_erf>(A.get_ref());
   }
@@ -1077,7 +1073,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_erfc> >::result
 erfc(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_erfc>(A);
   }
@@ -1090,7 +1086,7 @@ arma_inline
 typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_erfc> >::result
 erfc(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_erfc>(A.get_ref());
   }
@@ -1106,7 +1102,7 @@ arma_inline
 typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_lgamma> >::result
 lgamma(const T1& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOp<T1, eop_lgamma>(A);
   }
@@ -1119,13 +1115,42 @@ arma_inline
 typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_lgamma> >::result
 lgamma(const BaseCube<typename T1::elem_type,T1>& A)
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   return eOpCube<T1, eop_lgamma>(A.get_ref());
   }
 
 
 
+//
+// tgamma
+
+template<typename T1>
+arma_warn_unused
+arma_inline
+typename enable_if2< (is_arma_type<T1>::value && is_cx<typename T1::elem_type>::no), const eOp<T1, eop_tgamma> >::result
+tgamma(const T1& A)
+  {
+  arma_debug_sigprint();
+  
+  return eOp<T1, eop_tgamma>(A);
+  }
+
+
+
+template<typename T1>
+arma_warn_unused
+arma_inline
+typename enable_if2< is_cx<typename T1::elem_type>::no, const eOpCube<T1, eop_tgamma> >::result
+tgamma(const BaseCube<typename T1::elem_type,T1>& A)
+  {
+  arma_debug_sigprint();
+  
+  return eOpCube<T1, eop_tgamma>(A.get_ref());
+  }
+
+
+
 // the functions below are currently unused; reserved for potential future use
 
 template<typename T1> void exp_approx(const T1&) { arma_stop_logic_error("unimplemented"); }
